@@ -9,17 +9,105 @@ use crate::lua::Lua;
 use crate::string::String;
 use crate::table::Table;
 use crate::types::Integer;
-use crate::util::{assert_stack, protect_lua, StackGuard};
+use crate::util::{check_stack, protect_lua, StackGuard};
 use crate::value::{ToLua, Value};
 
 /// A struct for serializing Rust values into Lua values.
-pub struct Serializer<'lua>(pub &'lua Lua);
+#[derive(Debug)]
+pub struct Serializer<'lua> {
+    lua: &'lua Lua,
+    options: Options,
+}
+
+/// A struct with options to change default serializer behavior.
+#[derive(Debug, Clone, Copy)]
+#[non_exhaustive]
+pub struct Options {
+    /// If true, sequence serialization to a Lua table will create table
+    /// with the [`array_metatable`] attached.
+    ///
+    /// Default: **true**
+    ///
+    /// [`array_metatable`]: ../trait.LuaSerdeExt.html#tymethod.array_metatable
+    pub set_array_metatable: bool,
+
+    /// If true, serialize `None` (part of `Option` type) to [`null`].
+    /// Otherwise it will be set to Lua [`Nil`].
+    ///
+    /// Default: **true**
+    ///
+    /// [`null`]: ../trait.LuaSerdeExt.html#tymethod.null
+    /// [`Nil`]: ../../enum.Value.html#variant.Nil
+    pub serialize_none_to_null: bool,
+
+    /// If true, serialize `Unit` (type of `()` in Rust) and Unit structs to [`null`].
+    /// Otherwise it will be set to Lua [`Nil`].
+    ///
+    /// Default: **true**
+    ///
+    /// [`null`]: ../trait.LuaSerdeExt.html#tymethod.null
+    /// [`Nil`]: ../../enum.Value.html#variant.Nil
+    pub serialize_unit_to_null: bool,
+}
+
+impl Default for Options {
+    fn default() -> Self {
+        Options {
+            set_array_metatable: true,
+            serialize_none_to_null: true,
+            serialize_unit_to_null: true,
+        }
+    }
+}
+
+impl Options {
+    /// Retruns a new instance of `Options` with default parameters.
+    pub fn new() -> Self {
+        Self::default()
+    }
+
+    /// Sets [`set_array_metatable`] option.
+    ///
+    /// [`set_array_metatable`]: #structfield.set_array_metatable
+    pub fn set_array_metatable(mut self, enabled: bool) -> Self {
+        self.set_array_metatable = enabled;
+        self
+    }
+
+    /// Sets [`serialize_none_to_null`] option.
+    ///
+    /// [`serialize_none_to_null`]: #structfield.serialize_none_to_null
+    pub fn serialize_none_to_null(mut self, enabled: bool) -> Self {
+        self.serialize_none_to_null = enabled;
+        self
+    }
+
+    /// Sets [`serialize_unit_to_null`] option.
+    ///
+    /// [`serialize_unit_to_null`]: #structfield.serialize_unit_to_null
+    pub fn serialize_unit_to_null(mut self, enabled: bool) -> Self {
+        self.serialize_unit_to_null = enabled;
+        self
+    }
+}
+
+impl<'lua> Serializer<'lua> {
+    /// Creates a new Lua Serializer with default options.
+    pub fn new(lua: &'lua Lua) -> Self {
+        Self::new_with_options(lua, Options::default())
+    }
+
+    /// Creates a new Lua Serializer with custom options.
+    pub fn new_with_options(lua: &'lua Lua, options: Options) -> Self {
+        Serializer { lua, options }
+    }
+}
 
 macro_rules! lua_serialize_number {
     ($name:ident, $t:ty) => {
         #[inline]
         fn $name(self, value: $t) -> Result<Value<'lua>> {
-            value.to_lua(self.0)
+            value.to_lua(self.lua)
         }
     };
 }
@@ -62,35 +150,47 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
 
     #[inline]
     fn serialize_str(self, value: &str) -> Result<Value<'lua>> {
-        self.0.create_string(value).map(Value::String)
+        self.lua.create_string(value).map(Value::String)
     }
 
     #[inline]
     fn serialize_bytes(self, value: &[u8]) -> Result<Value<'lua>> {
-        self.0.create_string(value).map(Value::String)
+        self.lua.create_string(value).map(Value::String)
     }
 
     #[inline]
     fn serialize_none(self) -> Result<Value<'lua>> {
-        self.0.null()
+        if self.options.serialize_none_to_null {
+            Ok(self.lua.null())
+        } else {
+            Ok(Value::Nil)
+        }
     }
 
     #[inline]
     fn serialize_some<T>(self, value: &T) -> Result<Value<'lua>>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         value.serialize(self)
     }
 
     #[inline]
     fn serialize_unit(self) -> Result<Value<'lua>> {
-        self.0.null()
+        if self.options.serialize_unit_to_null {
+            Ok(self.lua.null())
+        } else {
+            Ok(Value::Nil)
+        }
     }
 
     #[inline]
     fn serialize_unit_struct(self, _name: &'static str) -> Result<Value<'lua>> {
-        self.0.null()
+        if self.options.serialize_unit_to_null {
+            Ok(self.lua.null())
+        } else {
+            Ok(Value::Nil)
+        }
     }
 
     #[inline]
@@ -106,7 +206,7 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
     #[inline]
     fn serialize_newtype_struct<T>(self, _name: &'static str, value: &T) -> Result<Value<'lua>>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         value.serialize(self)
     }
@@ -120,11 +220,11 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
         value: &T,
     ) -> Result<Value<'lua>>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
-        let table = self.0.create_table()?;
-        let variant = self.0.create_string(variant)?;
-        let value = self.0.to_value(value)?;
+        let table = self.lua.create_table()?;
+        let variant = self.lua.create_string(variant)?;
+        let value = self.lua.to_value_with(value, self.options)?;
         table.raw_set(variant, value)?;
         Ok(Value::Table(table))
     }
@@ -132,9 +232,12 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
     #[inline]
     fn serialize_seq(self, len: Option<usize>) -> Result<Self::SerializeSeq> {
         let len = len.unwrap_or(0) as c_int;
-        let table = self.0.create_table_with_capacity(len, 0)?;
-        table.set_metatable(Some(self.0.array_metatable()?));
-        Ok(SerializeVec { table })
+        let table = self.lua.create_table_with_capacity(len, 0)?;
+        if self.options.set_array_metatable {
+            table.set_metatable(Some(self.lua.array_metatable()));
+        }
+        let options = self.options;
+        Ok(SerializeVec { table, options })
     }
 
     #[inline]
@@ -159,9 +262,11 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
         variant: &'static str,
         _len: usize,
     ) -> Result<Self::SerializeTupleVariant> {
-        let name = self.0.create_string(variant)?;
-        let table = self.0.create_table()?;
-        Ok(SerializeTupleVariant { name, table })
+        Ok(SerializeTupleVariant {
+            name: self.lua.create_string(variant)?,
+            table: self.lua.create_table()?,
+            options: self.options,
+        })
     }
 
     #[inline]
@@ -169,7 +274,8 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
         let len = len.unwrap_or(0) as c_int;
         Ok(SerializeMap {
             key: None,
-            table: self.0.create_table_with_capacity(0, len)?,
+            table: self.lua.create_table_with_capacity(0, len)?,
+            options: self.options,
         })
     }
 
@@ -186,14 +292,18 @@ impl<'lua> ser::Serializer for Serializer<'lua> {
         variant: &'static str,
         len: usize,
     ) -> Result<Self::SerializeStructVariant> {
-        let name = self.0.create_string(variant)?;
-        let table = self.0.create_table_with_capacity(0, len as c_int)?;
-        Ok(SerializeStructVariant { name, table })
+        Ok(SerializeStructVariant {
+            name: self.lua.create_string(variant)?,
+            table: self.lua.create_table_with_capacity(0, len as c_int)?,
+            options: self.options,
+        })
     }
 }
 
+#[doc(hidden)]
 pub struct SerializeVec<'lua> {
     table: Table<'lua>,
+    options: Options,
 }
 
 impl<'lua> ser::SerializeSeq for SerializeVec<'lua> {
@@ -202,24 +312,20 @@ impl<'lua> ser::SerializeSeq for SerializeVec<'lua> {
 
     fn serialize_element<T>(&mut self, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         let lua = self.table.0.lua;
-        let value = lua.to_value(value)?;
+        let value = lua.to_value_with(value, self.options)?;
         unsafe {
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 4);
+            check_stack(lua.state, 5)?;
 
             lua.push_ref(&self.table.0);
             lua.push_value(value)?;
-
-            unsafe extern "C" fn push_to_table(state: *mut ffi::lua_State) -> c_int {
-                let len = ffi::lua_rawlen(state, -2) as Integer;
+            let len = ffi::lua_rawlen(lua.state, -2) as Integer;
+            protect_lua(lua.state, 2, 0, |state| {
                 ffi::lua_rawseti(state, -2, len + 1);
-                1
-            }
-
-            protect_lua(lua.state, 2, push_to_table)
+            })
         }
     }
 
@@ -234,7 +340,7 @@ impl<'lua> ser::SerializeTuple for SerializeVec<'lua> {
 
     fn serialize_element<T>(&mut self, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         ser::SerializeSeq::serialize_element(self, value)
     }
@@ -250,7 +356,7 @@ impl<'lua> ser::SerializeTupleStruct for SerializeVec<'lua> {
 
     fn serialize_field<T>(&mut self, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         ser::SerializeSeq::serialize_element(self, value)
     }
@@ -260,9 +366,11 @@ impl<'lua> ser::SerializeTupleStruct for SerializeVec<'lua> {
     }
 }
 
+#[doc(hidden)]
 pub struct SerializeTupleVariant<'lua> {
     name: String<'lua>,
     table: Table<'lua>,
+    options: Options,
 }
 
 impl<'lua> ser::SerializeTupleVariant for SerializeTupleVariant<'lua> {
@@ -271,11 +379,12 @@ impl<'lua> ser::SerializeTupleVariant for SerializeTupleVariant<'lua> {
 
     fn serialize_field<T>(&mut self, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         let lua = self.table.0.lua;
         let idx = self.table.raw_len() + 1;
-        self.table.raw_insert(idx, lua.to_value(value)?)
+        self.table
+            .raw_insert(idx, lua.to_value_with(value, self.options)?)
     }
 
     fn end(self) -> Result<Value<'lua>> {
@@ -286,9 +395,11 @@ impl<'lua> ser::SerializeTupleVariant for SerializeTupleVariant<'lua> {
     }
 }
 
+#[doc(hidden)]
 pub struct SerializeMap<'lua> {
     table: Table<'lua>,
     key: Option<Value<'lua>>,
+    options: Options,
 }
 
 impl<'lua> ser::SerializeMap for SerializeMap<'lua> {
@@ -297,23 +408,23 @@ impl<'lua> ser::SerializeMap for SerializeMap<'lua> {
 
     fn serialize_key<T>(&mut self, key: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         let lua = self.table.0.lua;
-        self.key = Some(lua.to_value(key)?);
+        self.key = Some(lua.to_value_with(key, self.options)?);
         Ok(())
     }
 
     fn serialize_value<T>(&mut self, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         let lua = self.table.0.lua;
         let key = mlua_expect!(
             self.key.take(),
             "serialize_value called before serialize_key"
         );
-        let value = lua.to_value(value)?;
+        let value = lua.to_value_with(value, self.options)?;
         self.table.raw_set(key, value)
     }
 
@@ -328,7 +439,7 @@ impl<'lua> ser::SerializeStruct for SerializeMap<'lua> {
 
     fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         ser::SerializeMap::serialize_key(self, key)?;
         ser::SerializeMap::serialize_value(self, value)
@@ -339,9 +450,11 @@ impl<'lua> ser::SerializeStruct for SerializeMap<'lua> {
     }
 }
 
+#[doc(hidden)]
 pub struct SerializeStructVariant<'lua> {
     name: String<'lua>,
     table: Table<'lua>,
+    options: Options,
 }
 
 impl<'lua> ser::SerializeStructVariant for SerializeStructVariant<'lua> {
@@ -350,10 +463,11 @@ impl<'lua> ser::SerializeStructVariant for SerializeStructVariant<'lua> {
 
     fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<()>
     where
-        T: ?Sized + Serialize,
+        T: Serialize + ?Sized,
     {
         let lua = self.table.0.lua;
-        self.table.raw_set(key, lua.to_value(value)?)?;
+        self.table
+            .raw_set(key, lua.to_value_with(value, self.options)?)?;
         Ok(())
     }
 
